/*
 * Copyright (C) 2009 Andre Noll <maan@systemlinux.org>
 *
 * Licensed under the GPL v2. For licencing details see COPYING.
 */

/** \file amp_filter.c Paraslash's amplify filter. */

#include "para.h"
#include "amp_filter.cmdline.h"
#include "list.h"
#include "sched.h"
#include "ggo.h"
#include "filter.h"
#include "string.h"
#include "error.h"

/** The size of the output data buffer. */
#define AMP_CHUNK_SIZE 40960

extern char *stat_item_values[NUM_STAT_ITEMS];

/** Data specific to the amplify filter. */
struct private_amp_data {
	/** Points to the configuration data for this instance of this filter. */
	struct amp_filter_args_info *conf;
	/** Amplification factor. */
	unsigned amp;
};

static ssize_t amp_convert(char *inbuf, size_t inbuf_len, struct filter_node *fn)
{
	size_t i, length = PARA_MIN((inbuf_len / 2) * 2,
		(fn->bufsize - fn->loaded) / 2 * 2);
	struct private_amp_data *pad = fn->private_data;
	int16_t *ip = (int16_t *)inbuf, *op = (int16_t *)(fn->buf + fn->loaded);

	if (!length)
		return 0;
	for (i = 0; i < length / 2; i++) {
		int x = (PARA_ABS(*ip) * (64 + pad->amp)) >> 6;
		*op++ = *ip++ > 0? PARA_MIN(x, 32767) : PARA_MAX(-x, -32768);
	}
	fn->loaded += length;
	return length;
}

static void amp_close(struct filter_node *fn)
{
	free(fn->private_data);
	free(fn->buf);
}

static int amp_parse_config(int argc, char **argv, void **config)
{
	struct amp_filter_args_info *amp_conf = para_calloc(sizeof(*amp_conf));
	int ret = -E_AMP_SYNTAX;

	if (amp_cmdline_parser(argc, argv, amp_conf))
		goto err;
	ret = -ERRNO_TO_PARA_ERROR(EINVAL);
	if (amp_conf->amp_arg < 0)
		goto err;
	PARA_NOTICE_LOG("amplification: %u (scaling factor: %1.2f)\n",
		amp_conf->amp_arg, amp_conf->amp_arg / 64.0 + 1.0);
	*config = amp_conf;
	return 1;
err:
	free(amp_conf);
	return ret;
}

static void amp_open(struct filter_node *fn)
{
	struct private_amp_data *pad = para_calloc(sizeof(*pad));

	pad->conf = fn->conf;
	fn->private_data = pad;
	if (!pad->conf->amp_given && stat_item_values[SI_AMPLIFICATION]) {
		int i = SI_AMPLIFICATION;
		char *s = stat_item_values[i] + strlen(status_item_list[i]) + 1;
		sscanf(s, "%u", &pad->amp);
	} else
		pad->amp = pad->conf->amp_arg;
	fn->bufsize = AMP_CHUNK_SIZE;
	fn->buf = para_malloc(fn->bufsize);
}

/**
 * The init function of the amplify filter.
 *
 * \param f Pointer to the struct to initialize.
 */
void amp_filter_init(struct filter *f)
{
	struct amp_filter_args_info dummy;

	amp_cmdline_parser_init(&dummy);
	f->open = amp_open;
	f->close = amp_close;
	f->convert = amp_convert;
	f->parse_config = amp_parse_config;
	f->help = (struct ggo_help) {
		.short_help = amp_filter_args_info_help,
		.detailed_help = amp_filter_args_info_detailed_help
	};
}
